/*  Monkey HTTP Daemon
 *  ------------------
 *  Copyright (C) 2001-2003, Eduardo Silva P.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <pthread.h>

#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <time.h>
#include <netdb.h>
#include <sys/wait.h>
#include <signal.h>

#include <arpa/inet.h> 
#include <netinet/in.h>
#include <sys/types.h>

#include "monkey.h"

#define MAX_TIMES 10000
#define POST_TIMEOUT 0

int Request_Main(struct request *s_request)
{

	int num_bytes=0, recv_timeout, times=0, status=0, limit_time;
	char remote_request[MAX_REQUEST_BODY];
	struct vhost *vhost;
	
	/* Valores iniciales */
	s_request->log->ip=strdup(PutIP(remote));
	s_request->log->datetime=M_strdup(PutTime());
	s_request->log->final_response=M_HTTP_OK;
	s_request->log->status=S_LOG_ON;
	s_request->status=VAR_ON;
	s_request->method=METHOD_NOT_FOUND;
	s_request->forcegetdir=VAR_OFF;

	s_request->headers->range_values[0]=-1;
	s_request->headers->range_values[1]=-1;
	
	if(s_request->counter_connections>0)
		recv_timeout=config->keep_alive_timeout; 
	else
		recv_timeout=config->timeout;

	memset(remote_request, '\0', sizeof(remote_request));
	limit_time = (int) time(NULL) + recv_timeout;
	
	/* Recibiendo Peticion */
	do {		
		times++;

		num_bytes=Socket_Timeout(s_request->socket, remote_request+strlen(remote_request), \
								MAX_REQUEST_BODY - strlen(remote_request) - 1, recv_timeout, ST_RECV);

		if((int) time(NULL) >= limit_time){
			Request_Error(M_CLIENT_REQUEST_TIMEOUT, s_request, 1, s_request->log);
			return -1;
		}
		
		/* Timeout ? */
		if(num_bytes==-2){
			if(s_request->counter_connections > 0){ /* es conexion persistente ? */
				return 2;    /* Exit por Timeout de conexion persistente */
			}
			else{
				if(recv_timeout==config->timeout){
					Request_Error(M_CLIENT_REQUEST_TIMEOUT, s_request, 1,s_request->log);
					return -1; /* Exit por Timeout normal */
				}
				else
					continue;
			}
		}
		/* Exit de conexion persistente */
		if(num_bytes<=0){
			return -1;
		}

		if(times==1) {
			 /* Validando formato de la primera cabecera */
			int i, count=0;
			for(i=0; i<strlen(remote_request) && count<2; i++){
				if(remote_request[i]==' ') count++;
			}
			if(count<2){
				Request_Error(M_CLIENT_BAD_REQUEST, s_request,1,s_request->log);
				return -1;
			}
			if(Get_method_from_request(remote_request)==POST_METHOD)
				recv_timeout=POST_TIMEOUT;
		}
	} while(!strstr(remote_request,"\r\n\r\n") && !strstr(remote_request,"\n\n") && times<RECV_MAX_TIMES);

	s_request = (struct request *) Request_Strip_Header(s_request, remote_request);

	s_request->user_home=VAR_OFF;
	strncpy(s_request->temp_path, config->server_root, MAX_PATH);
	s_request->temp_path[MAX_PATH - 1]='\0';
	s_request->counter_connections++;
	
	/* IP o URL permitida? */
	if(Deny_Check(s_request->uri_processed)==-1 || Deny_Check(s_request->query_string)){
		s_request->log->final_response=M_CLIENT_FORBIDDEN;
		Request_Error(M_CLIENT_FORBIDDEN, s_request,1,s_request->log);
		return -1;
	}
	
	/* HTTP/1.1 requiere de especificacion del Host */
	if(!s_request->host && s_request->protocol==HTTP_11){
		s_request->log->final_response=M_CLIENT_BAD_REQUEST;
		Request_Error(M_CLIENT_BAD_REQUEST, s_request,1,s_request->log);
		return -1;
	}
	
	/* Host vacio (HTTP/1.0) */
	if(!s_request->host){
		s_request->host=config->servername;
	}

	/* Server Signature */
	if(config->hideversion==VAR_OFF){
		config->server_signature = m_build_buffer("Monkey/%s Server (Host: %s, Port: %i)", 
			VERSION, s_request->host, config->serverport);
	}
	else{
		config->server_signature = m_build_buffer("Monkey Server (Host: %s, Port: %i)", 
			s_request->host, config->serverport);
	}
	
	/* Asignando info: SERVER_SCRIPTALIAS */
	if(config->server_scriptalias!=NULL){
		s_request->scriptalias = (char **) alloca(sizeof(char *) * 3);
		s_request->scriptalias[0] = config->server_scriptalias[0];
		s_request->scriptalias[1] = config->server_scriptalias[1];
		s_request->scriptalias[2]='\0';
	}
		
	/* Metodo no permitido ? */
	if(s_request->method==METHOD_NOT_ALLOWED){
		s_request->log->final_response=M_CLIENT_METHOD_NOT_ALLOWED;
		Request_Error(M_CLIENT_METHOD_NOT_ALLOWED, s_request,1,s_request->log);
		return -1;
	}

	/* Validando version del protocolo */
	if(!strstr(PROTOCOLS, get_name_protocol(s_request->protocol))) {
		s_request->log->final_response=M_SERVER_HTTP_VERSION_UNSUP;
		Request_Error(M_SERVER_HTTP_VERSION_UNSUP, s_request,1,s_request->log);
		return -1;
	}
  
  
	/*  Peticion a un VirtualHost ? */
	if((vhost=VHOST_Find(s_request->host))!=NULL) {
		strncpy(s_request->temp_path,vhost->documentroot,MAX_PATH);
		s_request->temp_path[MAX_PATH - 1]='\0';

		s_request->scriptalias = (char **) alloca(sizeof(char *) * 3);
		s_request->scriptalias[0]=vhost->cgi_alias;
		s_request->scriptalias[1]=vhost->cgi_path;
		s_request->scriptalias[2]='\0';
		s_request->forcegetdir=vhost->forcegetdir;
	}

	/*  Peticion CGI ? */
	if(config->server_scriptalias!=NULL){
		if((strncmp(config->server_scriptalias[0], s_request->uri_processed, strlen(config->server_scriptalias[0])))==0){
			int cgi_status;
			cgi_status=M_CGI_main(s_request, s_request->log, remote_request);
			/* Codes:
		-1 : Fallo de permisos
		-2 : Timeout
		-3 : Internal Server Error
	*/
			if(cgi_status==M_CGI_TIMEOUT || cgi_status==M_CGI_INTERNAL_SERVER_ERR){
				Request_Error(s_request->log->final_response, s_request, 1, s_request->log);	
			}
			return cgi_status;
		}
	}

	/*  Peticion a un directorio de usuario ? */
	if(strncmp(s_request->uri_processed, USER_HOME_STRING, strlen(USER_HOME_STRING))==0 &&
		config->user_dir){
		if(User_main(s_request)!=0)
			return -1;
	}

	/* Verificacion del Metodo */
	if(s_request->method==GET_METHOD || s_request->method==HEAD_METHOD){
			status=M_METHOD_Get_and_Head(s_request);
	}
	else {
		if(s_request->method==POST_METHOD){
			if((status=M_METHOD_Post(s_request, remote_request))==-1){
				M_free(s_request->post_variables);
				return status;
			}
			status=M_METHOD_Get_and_Head(s_request);
		}
	}

	return status;
}

/* Retorna el metodo del request:
 Utilizamos esta funcion para conocer
 el metodo antes de que termine el envio 
 del request */
int Get_method_from_request(char *request)
{
	char *method, *last, first_line[1024];
	
	memset(first_line,'\0',sizeof(first_line));
	strncpy(first_line, request, 1024);
	first_line[1023]='\0';
	method=strtok_r(first_line," ", &last);

	return M_METHOD_get_number(method);
}

/* Retorna una estructura con el Metodo, URI del archivo y
version del Protocolo que ha sido enviado en el request */
struct request *Request_Strip_Header(struct request *sr, char *request_body)
{
	char *aux=0, *last=0, *uri_aux=0;
	char aux_body[MAX_REQUEST_BODY];
	
	memset(aux_body, '\0', sizeof(aux_body));
	strncpy(aux_body,request_body,MAX_REQUEST_BODY);
	aux_body[MAX_REQUEST_BODY -1]='\0';

	/* Method */
	if((aux=strtok_r(aux_body, " ", &last))!=NULL)
		sr->method=M_METHOD_get_number(aux);
	else
		sr->method=METHOD_NOT_FOUND;

        /* Request URI */
	if((aux=strtok_r(NULL," ", &last))!=NULL){
		uri_aux = strdup(aux);
	}

	/* HTTP Version */
	if((aux=strtok_r(NULL,"\r", &last))!=NULL){
		sr->protocol=get_version_protocol(aux);
	}

	last=0;
	sr->uri = strdup(strtok_r(uri_aux,"?", &last));
	
	/* Query string */	
	if((aux=strtok_r(NULL,"\r", &last))!=NULL){
		sr->query_string=M_strdup(aux);
	}

	M_free(uri_aux);
	
	/* URI processed */
	sr->uri_processed = m_build_buffer("%s", get_real_string( sr->uri ));

	strncpy(aux_body,request_body,MAX_REQUEST_BODY);
	aux_body[MAX_REQUEST_BODY -1]='\0';
	
	/* Validando el Host */
	if((strstr2(aux_body, RH_HOST))!=NULL){
		sr->host = Request_Find_Variable(aux_body, RH_HOST);
		/* El host es de la forma xxxxx:yy ???? */
		if(strstr(sr->host, ":") != NULL ){
			last=0;
			sr->host=M_strdup(strtok_r(sr->host, ":", &last));
			sr->port=atoi(strtok_r(NULL,"\n", &last));
		}
		else
			sr->port=config->standard_port;
	}
	else
		sr->host=NULL;
		
	/* Variables generales del header remoto */
	sr->keep_alive=VAR_OFF;
	if((strstr2(aux_body, RH_CONNECTION))!=NULL && sr->protocol==HTTP_11){
		sr->connection = Request_Find_Variable(request_body, RH_CONNECTION);
		if((strstr2(sr->connection,"Keep-Alive"))!=NULL){
			sr->keep_alive=VAR_ON;
		}
	}
	
	if((strstr2(aux_body, RH_ACCEPT))!=NULL)
		sr->accept = Request_Find_Variable(request_body, RH_ACCEPT);
		
	if((strstr2(aux_body, RH_ACCEPT_CHARSET))!=NULL)
		sr->accept_charset = Request_Find_Variable(request_body, RH_ACCEPT_CHARSET);

	if((strstr2(request_body, RH_ACCEPT_ENCODING))!=NULL)
		sr->accept_encoding = Request_Find_Variable(request_body, RH_ACCEPT_ENCODING);
	
	if((strstr2(request_body, RH_ACCEPT_LANGUAGE))!=NULL)
		sr->accept_language = Request_Find_Variable(request_body, RH_ACCEPT_LANGUAGE);

	if((strstr2(request_body, RH_COOKIE))!=NULL)
		sr->cookies = Request_Find_Variable(request_body, RH_COOKIE);

	if((strstr2(request_body, RH_REFERER))!=NULL)
		sr->referer = Request_Find_Variable(request_body, RH_REFERER);

	if((strstr2(aux_body, RH_USER_AGENT))!=NULL)
		sr->user_agent = Request_Find_Variable(request_body, RH_USER_AGENT);
		
	if((strstr2(request_body, RH_RANGE))!=NULL)
		sr->range = Request_Find_Variable(request_body, RH_RANGE);
	
	return (struct request *) sr;
}

/* Devuelve valor de variables enviadas en el request */
char *Request_Find_Variable(char *request_body,  char *string)
{
	char *buffer=0, *last=0, aux_body[MAX_REQUEST_BODY];
	
	strncpy(aux_body,request_body,MAX_REQUEST_BODY);
	aux_body[MAX_REQUEST_BODY -1]='\0';

	/* Existe *string en request_body ??? */	
	if ((buffer = (char *) strstr2(aux_body, string)) == NULL)
		return NULL;

	strncpy(aux_body, buffer, MAX_REQUEST_BODY);
	if((buffer=strtok_r(aux_body, " ", &last))==NULL){	/* Leer la variable */
		return NULL;
	}
	
	if((buffer=strtok_r(NULL, "\n", &last))==NULL){	/* Leer el contenido */
		return "";
	}

	buffer[strlen(buffer)-1] = '\0';
	return (char *) strdup(buffer);
}

/* Encuentra index.xxx para el request */
int FindIndex(char *pathfile)
{
	char fileaux[255];

	struct indexfile *aux_index;
	
	aux_index=first_index;

	while(aux_index!=NULL) {
		snprintf(fileaux,MAX_PATH,"%s/%s",pathfile,aux_index->indexname);
		if(access(fileaux,F_OK)==0) {
			memset(pathfile,'\0',sizeof(pathfile));
			strncpy(pathfile,fileaux,MAX_PATH);
			pathfile[MAX_PATH - 1]='\0';
			return 0;
		}
		aux_index=aux_index->next;
	}

	return -1;
}

/* Manejo de errores en peticiones */
void Request_Error(int num_error, struct request *s_request, int debug, struct log_info *s_log)
{
	int max_length=255;
	char *page_default=0, aux_message[max_length];
			
	if(!s_log) 
		s_log=malloc(sizeof(struct log_info));
		
	switch(num_error) {
		case M_CLIENT_BAD_REQUEST:
			fdprintf(s_request->socket, NO_CHUNKED, "HTTP/1.1 400 Bad Request\r\n");
			page_default=Set_Page_Default(max_length, "Bad Request", "", config->server_signature);
			s_log->error_msg=m_build_buffer("[error 400] Bad request");
			break;

		case M_CLIENT_FORBIDDEN:
			fdprintf(s_request->socket, NO_CHUNKED, "HTTP/1.1 403 Forbidden\r\n");
			page_default=Set_Page_Default(max_length, "Forbidden", s_request->uri, config->server_signature);
			s_log->error_msg=m_build_buffer("[error 403] Forbidden %s",s_request->uri);
			break;

		case M_CLIENT_NOT_FOUND:
			fdprintf(s_request->socket, NO_CHUNKED, "HTTP/1.1 404 Not Found\r\n");
			snprintf(aux_message, max_length, "The requested URL %.100s  was not found on this server.", (char *) s_request->uri);
			page_default=Set_Page_Default(max_length, "Not Found", aux_message, config->server_signature);
			s_log->error_msg=m_build_buffer("[error 404] Not Found %s",s_request->uri);
			break;

		case M_CLIENT_METHOD_NOT_ALLOWED:
			fdprintf(s_request->socket, NO_CHUNKED, "HTTP/1.1 405 Method Not Allowed\r\n");
			page_default=Set_Page_Default(max_length, "Method Not Allowed",s_request->uri, config->server_signature);
			s_log->final_response=M_CLIENT_METHOD_NOT_ALLOWED;
			s_log->error_msg=m_build_buffer("[error 405] Method Not Allowed %s", M_METHOD_get_name(s_request->method));
			break;

		case M_CLIENT_REQUEST_TIMEOUT:
			fdprintf(s_request->socket, NO_CHUNKED, "HTTP/1.1 408 Request Timeout\r\n");
			s_log->status=S_LOG_OFF;
			s_log->error_msg=m_build_buffer("[error 408] Request Timeout");
			break;

		case M_CLIENT_LENGHT_REQUIRED:
			fdprintf(s_request->socket, NO_CHUNKED, "HTTP/1.1 411 Length Required\r\n");
			s_log->error_msg=m_build_buffer("[error 411] Length Required");
			break;
			
		case M_SERVER_INTERNAL_ERROR:
			fdprintf(s_request->socket, NO_CHUNKED, "HTTP/1.1 500 Internal Server Error\r\n");
			snprintf(aux_message, max_length, "Problems found running %s ",s_request->uri);
			page_default=Set_Page_Default(max_length, "Internal Server Error",aux_message, config->server_signature);
			s_log->error_msg=m_build_buffer("[error 411] Internal Server Error %s",s_request->uri);
			break;
			
		case M_SERVER_HTTP_VERSION_UNSUP:
			fdprintf(s_request->socket, NO_CHUNKED, "HTTP/1.1 505 HTTP Version Not Supported\r\n");
			page_default=Set_Page_Default(max_length, "HTTP Version Not Supported"," ",config->server_signature);
			s_log->error_msg=m_build_buffer("[error 505] HTTP Version Not Supported");
			break;
	}

	s_log->final_response=num_error;
	
	s_request->headers->status = 0;
	s_request->headers->content_length = 0;
	s_request->headers->location = NULL ;
	s_request->headers->cgi = SH_NOCGI;
	s_request->headers->pconnections_left = 0;
	
	if(!page_default)
		s_request->headers->content_type = NULL;
	else
		s_request->headers->content_type = m_build_buffer("text/html");

	M_METHOD_send_headers(s_request->socket, s_request->headers, s_log);
	
	if(debug==1) 
		fdprintf(s_request->socket, NO_CHUNKED, "%s", page_default);
}

/* Genera pagina de error */
char *Set_Page_Default(int max_length, char *title, char *message, char *signature)
{
	char *page=0;

	page=m_build_buffer("<HTML><BODY><H1>%s</H1>%s<BR><HR> \
		<ADDRESS>%s</ADDRESS></BODY></HTML>", title, message, signature);
	return (char *) page;
}

/* Establece Timeout para send() y recv() */
int Socket_Timeout(int s, char *buf, int len, int timeout, int recv_send)
{
	fd_set fds;
	int n=0, status;
	struct timeval tv;

	FD_ZERO(&fds);
	FD_SET(s,&fds);
	
	tv.tv_sec=timeout;
	tv.tv_usec=0;
	
	if(recv_send==ST_RECV)
		n=select(s+1,&fds,NULL,NULL,&tv); /* recv */
	else{
		n=select(s+1,NULL,&fds,NULL,&tv); /* send */
	}

	switch(n){
		case 0:
				return -2;
				break;
		case -1:
				kill(pthread_self(), SIGPIPE);
				return -1;
	}
	
	if(recv_send==ST_RECV)
		status=recv(s,buf,len, 0);
	else
		status=send(s,buf,len, 0);

	if( status < 0 ){
		kill(pthread_self(), SIGPIPE);
	}
	
	return status;
}
